package ch.usi.inf.sa4.sanmarinoes.smarthut.controller;

import ch.usi.inf.sa4.sanmarinoes.smarthut.error.NotFoundException;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.*;
import java.util.Set;

/**
 * An abstract controller for an input device that has output connected to it. Aids to create the
 * output add and output remove route
 *
 * @param <I> the type of device this controller is for
 * @param <O> the output device attached to I
 */
public abstract class InputDeviceConnectionController<
        I extends InputDevice, O extends OutputDevice> {

    private class IOPair {
        private final I input;
        private final O output;

        private IOPair(I input, O output) {
            this.input = input;
            this.output = output;
        }
    }

    private DeviceRepository<I> inputRepository;

    private DeviceRepository<O> outputReposiory;

    private Connector<I, O> connector;

    /**
     * Contstructs the controller by requiring essential object for the controller implementation
     *
     * @param inputRepository the input device repository
     * @param outputRepository the output device repository
     * @param connector a appropriate Connector instance for the I and O tyoes.
     */
    protected InputDeviceConnectionController(
            DeviceRepository<I> inputRepository,
            DeviceRepository<O> outputRepository,
            Connector<I, O> connector) {
        this.inputRepository = inputRepository;
        this.outputReposiory = outputRepository;
        this.connector = connector;
    }

    private IOPair checkConnectionIDs(Long inputId, Long outputId) throws NotFoundException {
        final I input =
                inputRepository
                        .findById(inputId)
                        .orElseThrow(() -> new NotFoundException("input device"));
        final O output =
                outputReposiory
                        .findById(outputId)
                        .orElseThrow(() -> new NotFoundException("output device"));
        return new IOPair(input, output);
    }

    /**
     * Implements the output device connection creation (add) route
     *
     * @param inputId input device id
     * @param outputId output device id
     * @return the list of output devices attached to the input device of id inputId
     * @throws NotFoundException if inputId or outputId are not valid
     */
    protected Set<? extends OutputDevice> addOutput(Long inputId, Long outputId)
            throws NotFoundException {
        final IOPair pair = checkConnectionIDs(inputId, outputId);
        connector.connect(pair.input, pair.output, true);
        outputReposiory.save(pair.output);
        return pair.input.getOutputs();
    }

    /**
     * Implements the output device connection destruction (remove) route
     *
     * @param inputId input device id
     * @param outputId output device id
     * @return the list of output devices attached to the input device of id inputId
     * @throws NotFoundException if inputId or outputId are not valid
     */
    protected Set<? extends OutputDevice> removeOutput(Long inputId, Long outputId)
            throws NotFoundException {
        final IOPair pair = checkConnectionIDs(inputId, outputId);
        connector.connect(pair.input, pair.output, false);
        outputReposiory.save(pair.output);
        return pair.input.getOutputs();
    }
}
