package ch.usi.inf.sa4.sanmarinoes.smarthut.models;

import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.Function;

/**
 * A rule on how to connect an input device type to an output device type
 *
 * @param <I> the input device type
 * @param <O> the output device type
 */
@FunctionalInterface
public interface Connector<I extends InputDevice, O extends OutputDevice> {

    /**
     * Connects or disconnects input to output
     *
     * @param input the input device
     * @param output the output device
     * @param connect true if connection, false if disconnection
     */
    void connect(I input, O output, boolean connect);

    /**
     * Produces a basic implementation of a connector, assuming there is a OneToMany relationship
     * between J and K
     *
     * @param outputsGetter the getter method of the set of outputs on the input class
     * @param inputSetter the setter method for the input id on the output class
     * @param <J> the input device type
     * @param <K> the output device type
     * @return a Connector implementation for the pair of types J and K
     */
    static <J extends InputDevice, K extends OutputDevice> Connector<J, K> basic(
            Function<J, Set<? super K>> outputsGetter, BiConsumer<K, Long> inputSetter) {
        return (i, o, connect) -> {
            if (connect) {
                outputsGetter.apply(i).add(o);
            } else {
                outputsGetter.apply(i).remove(o);
            }

            inputSetter.accept(o, connect ? i.getId() : null);
        };
    }
}
