package ch.usi.inf.sa4.sanmarinoes.smarthut.socket;


import ch.usi.inf.sa4.sanmarinoes.smarthut.config.GsonConfig;
import ch.usi.inf.sa4.sanmarinoes.smarthut.models.User;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import com.google.common.collect.Multimaps;
import com.google.gson.Gson;
import java.io.IOException;
import java.util.*;
import javax.websocket.*;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/** Endpoint of socket at URL /sensor-socket used to update the client with sensor information */
@Component
public class SensorSocketEndpoint extends Endpoint {

    private Gson gson = GsonConfig.gson();

    private AuthenticationMessageListener authenticationMessageListener;

    private Set<Session> unauthorizedClients = Collections.synchronizedSet(new HashSet<>());

    private Multimap<User, Session> authorizedClients =
            Multimaps.synchronizedMultimap(HashMultimap.create());

    @Autowired
    public SensorSocketEndpoint(AuthenticationMessageListener authenticationMessageListener) {
        this.authenticationMessageListener = authenticationMessageListener;
    }

    /**
     * Returns a synchronized set of socket sessions not yet authorized with a token
     *
     * @return a synchronized set of socket sessions not yet authorized with a token
     */
    public Set<Session> getUnauthorizedClients() {
        return unauthorizedClients;
    }

    /**
     * Returns a synchronized User to Session multimap with authorized sessions
     *
     * @return a synchronized User to Session multimap with authorized sessions
     */
    public Multimap<User, Session> getAuthorizedClients() {
        return authorizedClients;
    }

    /**
     * Given a message and a user, broadcasts that message in json to all associated clients and
     * returns the number of successful transfers
     *
     * @param message the message to send
     * @param u the user to which to send the message
     * @return number of successful transfer
     */
    public void broadcast(Object message, User u) {
        final HashSet<Session> sessions = new HashSet<>(authorizedClients.get(u));
        for (Session s : sessions) {
            try {
                if (s.isOpen()) {
                    s.getBasicRemote().sendText(gson.toJson(message));
                } else {
                    authorizedClients.remove(u, s);
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     * Handles the opening of a socket session with a client
     *
     * @param session the newly born session
     * @param config endpoint configuration
     */
    @Override
    public void onOpen(Session session, EndpointConfig config) {
        unauthorizedClients.add(session);
        session.addMessageHandler(
                authenticationMessageListener.newHandler(
                        session,
                        (u, s) -> {
                            unauthorizedClients.remove(s);
                            authorizedClients.put(u, s);
                        }));
    }
}
